package no.ntnu.idatt1002.demo.controller;

import java.time.LocalDate;
import java.util.List;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.chart.PieChart;
import javafx.scene.chart.PieChart.Data;
import javafx.scene.control.*;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.text.Text;
import javafx.stage.Modality;
import javafx.stage.Screen;
import javafx.stage.Stage;
import no.ntnu.idatt1002.demo.data.Budget.BudgetItem;
import no.ntnu.idatt1002.demo.data.Budget.BudgetRegister;
import no.ntnu.idatt1002.demo.data.Budget.FileHandlingBudget;
import no.ntnu.idatt1002.demo.data.Budget.FileHandlingBudgetArchive;
import no.ntnu.idatt1002.demo.data.Budget.FileHandlingSelectedBudget;
import no.ntnu.idatt1002.demo.data.Budget.GeneralBudget;
import no.ntnu.idatt1002.demo.data.Economics.ExpenseCategory;

import java.io.IOException;
import java.util.Optional;

/**
 * Controller for budget scene in the application. This controller manages all actions that relates to the budget tableview (add, edit and delete), the switching
 * of scenes from the budget scene to another scene, and the saving of the table, whenever the user switches to another scene.
 *
 * @author Anders Emil Bergan
 * @since 24.3.2023
 */
public class BudgetController extends FinanceController {
    private GeneralBudget general;

    @FXML
    private Button addBtn;

    @FXML
    private Button editBtn;

    @FXML
    private Button deleteBtn;

    @FXML
    private Button returnToMainMenuBtn;

    @FXML
    private Button backBtn;

    @FXML
    private Button continueBtn;

    @FXML
    private TableColumn<BudgetItem, Double> amountCol;

    @FXML
    private TableView<BudgetItem> budgetTableView = new TableView<>();

    @FXML
    private TableColumn<BudgetItem, ExpenseCategory> categoryCol;

    @FXML
    private TableColumn<BudgetItem, String> descriptionCol;

    @FXML
    private Text sum;

    @FXML
    private DatePicker date;

    @FXML
    private ObservableList<BudgetItem> budgetList;

    @FXML
    private PieChart budgetPieChart;

    @FXML
    private Label daysLeftLbl;

    BudgetRegister budgetRegister = new BudgetRegister();


    /**
     * Initializes the budget register, the observable budget list and the tableview, along with the values of the dropbox used for filtering the tableview.
     */

    @FXML
    public void initialize() {
        //TODO specify error messgage for when amount is exceeded / duplicate exists
        //todo properly close screen so things are saved
        //Initialize table columns
        categoryCol.setCellValueFactory(new PropertyValueFactory<BudgetItem, ExpenseCategory>("budgetCategory"));
        amountCol.setCellValueFactory(new PropertyValueFactory<BudgetItem, Double>("budgetAmount"));
        descriptionCol.setCellValueFactory(new PropertyValueFactory<BudgetItem, String>("budgetDescription"));

        try {
            System.out.println("Just in budget try");
            general = loadBudgetDataFromFile("Budget");
            budgetList = FXCollections.observableArrayList(general.getBudgetItems());
            budgetTableView.setItems(budgetList);
            System.out.println("After loading general");

            if (FileHandlingBudgetArchive.isBudgetRegisterEmpty()) {
                budgetRegister = new BudgetRegister();
            } else {
                budgetRegister = FileHandlingBudgetArchive.readBudgetArchive("");
            }

            if (FileHandlingBudget.isNewBudget("Budget")) {
                returnToMainMenuBtn.setOpacity(0);
                returnToMainMenuBtn.setDisable(true);
                daysLeftLbl.setOpacity(0);
            } else {
                refreshPieChart();
                //addBtn.setDisable(true);
                //editBtn.setDisable(true);
                //deleteBtn.setDisable(true);
                backBtn.setDisable(true);
                continueBtn.setDisable(true);
                backBtn.setOpacity(0);
                continueBtn.setOpacity(0);
            }
        } catch(IOException ioe) {
            ioe.printStackTrace();
            showErrorDialogBox(ioe.getMessage(), ioe.getMessage(), ioe.getMessage());
        }
        formatDatePicker();
    }

    public ObservableList<PieChart.Data> createPieChart() throws IllegalArgumentException { //
        ObservableList<PieChart.Data> budgetData = FXCollections.observableArrayList();
        List<ExpenseCategory> chosenCategories = general.getChosenBudgetCategories();

        //Only adds the budget data for chosen categories to the pie chart
        for (ExpenseCategory category : chosenCategories) {
            budgetData.add(new Data(category.toString().substring(0, 1).toUpperCase().
                concat(category.toString().substring(1)),
                general.getBudgetItem(category).getBudgetAmount()));
        }
        return budgetData;
    }

    /*private ObservableList<PieChart.Data> createBudgetPieChart() {
        return FXCollections.observableArrayList(
            new Data("Food", general.getBudgetItem(ExpenseCategory.FOOD).getBudgetAmount()),
            new Data("Books", general.getBudgetItem(ExpenseCategory.BOOKS).getBudgetAmount()),
            new Data("Clothes", general.getBudgetItem(ExpenseCategory.CLOTHES).getBudgetAmount()),
            new Data("Other", general.getBudgetItem(ExpenseCategory.OTHER).getBudgetAmount())
        );
    }*/

    @Override
    public void refreshPieChart() {
        this.budgetPieChart.setData(createPieChart());
    }

    /**
     * Method for disabling the date picker, yet having its opacity at max.
     */
    @Override
    public void formatDatePicker() {
        date.setValue(LocalDate.now());
        date.setDisable(true);
        date.setStyle("-fx-opacity: 1");
        date.getEditor().setStyle("-fx-opacity: 1");
    }

    @Override
    public void handleAddBtn(ActionEvent event) {
        handleEditBtn(event);
    }
    /**
     * Adds or edits a budget item, depending on what mode the DialogMode enum is at. The method brings up a dialog box popup in which the user can fill and choose
     * values that the budget item will have. Open exiting the popup, the changes the saved to the tableview.
     * @param event A button click on either the add or delete button.
     */
    @Override
    public void handleEditBtn(ActionEvent event) {
        BudgetItem item = null;
        String dialogTitle = "";
        DialogMode dialogMode;

        FXMLLoader loader = new FXMLLoader(SceneController.class.getResource("/view/AddBudgetNew.fxml"));
        Dialog<BudgetItem> dialog = new Dialog<>();
        dialog.initModality(Modality.APPLICATION_MODAL);

        //Try to load the FXML file onto another dialogPane
        try{
            dialog.getDialogPane().setContent(loader.load());
        } catch(IOException e) {
            showErrorDialogBox("Loading error", "Error in loading dialog", "An error occurred"
                + "when loading the AddBudget window");
        }

        //Loads the controller for the dialog box that appears whenever one adds or edits a budget item
        AddBudgetController budgetController = loader.getController();

        //Sets the title of the dialog box
        if(event.getSource().equals(addBtn)){
            dialogMode = DialogMode.ADD;
            dialogTitle = "New Budget";
        }
        else if (event.getSource().equals(editBtn) && budgetTableView.getSelectionModel().getSelectedItem() != null) {
            dialogMode = DialogMode.EDIT;
            dialogTitle = "Edit expense";
            //Gets the selected item from the table
            item = budgetTableView.getSelectionModel().getSelectedItem();
            //Binds the selected item to another item which is defined in the budgetcontroller
            budgetController.setBudget(item);
        } else {
            return;
        }

        dialog.setTitle(dialogTitle);
        // Show the Dialog and wait for the user to close it
        dialog.showAndWait();

        //Adds the new item to the register
        item = budgetController.getNewBudgetItem();
        if(item != null && dialogMode == DialogMode.ADD){
            try {
                general.addToBudgetBudgetItem(item);
            } catch(IllegalArgumentException e) {
                showErrorDialogBox(e.getMessage(), e.getMessage(), e.getMessage());
            }
        }
        //Updates the tableview using the register
        refreshTableView();
        refreshPieChart();
    }


    /**
     * Deletes an entry from the tableview, if an entry has been selected. The method brings up a popup window, asking for confirmation for deleting the entry.
     * @param event A button click on the delete button
     */
    @FXML
    public void handleDeleteBtn(ActionEvent event) {
        //Gets the selected item from the tableview
        BudgetItem item = budgetTableView.getSelectionModel().getSelectedItem();
        //Exits the method if nothing is selected
        if (item == null) {
            return;
        }
        //Brings up a confirmation popup
        String title = "Confirm Delete" ;
        String header = "Delete Confirmation";
        String content = "Are you sure you would like to delete the selected entry?";
        Optional<ButtonType> isConfirmed = showConfirmationDialog(title, header, content);
        if (isConfirmed.isPresent() && isConfirmed.get() == ButtonType.OK) {
            general.deleteItemFromBudget(item.getBudgetCategory());
            refreshTableView();
            refreshPieChart();
        }
    }

    /**
     * Method for synching the register with the tableview. The observable list to which the tableview is set, is being refilled with all the entries
     * in the register, keeping it updated with new changes.
     */
    @Override
    public void refreshTableView(){
        this.budgetList.setAll(general.getBudgetItems());
        //Refreshing the sum of the amounts of the budget
        //this.sum.setText(String.valueOf(general.totalSum()));
    }

    /**
     * Saves the changes made to the tableview by writing the information to a file.
     *
     * @throws IOException If an error occurs while writing to the file.
     */
    @Override
    public void saveDataToFile() throws IOException {
        FileHandlingBudget.writeGeneralBudgetToFile("Budget", general);
    }

    public void updateBudgetRegister(String budgetFolderName) {
        try {
            budgetRegister.addBudgetName(budgetFolderName);
            FileHandlingBudgetArchive.writeBudgetRegisterToArchive(budgetRegister);
        } catch(IOException ioe) {
            showErrorDialogBox(ioe.getMessage(), ioe.getMessage(), ioe.getMessage());
        }

    }

    /**
     * Switches scene, by loading a new FXML file and setting the scene to this location.
     * @param event A button click on the return to main menu button
     */
    @Override
    public void switchScene(ActionEvent event) {
        FXMLLoader loader = new FXMLLoader();
        try {
            if (event.getSource() == returnToMainMenuBtn) {
                //Adds unused categories to the table
                general.addUnusedCategories();
                loader.setLocation(getClass().getResource("/view/MainMenuNew.fxml"));
                //Always saving the data when switching scenes
                saveDataToFile();
            } else if (event.getSource() == continueBtn) {
                general.addUnusedCategories();
                updateBudgetRegister(FileHandlingSelectedBudget.readSelectedBudget());
                loader.setLocation(getClass().getResource("/view/MainMenuNew.fxml"));
                //Always saving the data when switching scenes
                saveDataToFile();
            } else if (event.getSource() == backBtn) {
                loader.setLocation(getClass().getResource("/view/dualList.fxml"));
            }
            Parent root = loader.load();
            Stage stage = (Stage) ((Node) event.getSource()).getScene().getWindow();
            Scene scene = new Scene(root);
            stage.setScene(scene);

            //Centralize the new screen
            Rectangle2D primScreenBounds = Screen.getPrimary().getVisualBounds();
            stage.setX((primScreenBounds.getWidth() - stage.getWidth()) / 2);
            stage.setY((primScreenBounds.getHeight() - stage.getHeight()) / 2);

            stage.show();

        } catch(Exception ioe) {
            ioe.printStackTrace();
            showErrorDialogBox("Loading error", "Error in loading", "Could not load"
                + " FXML file in: " + loader.getLocation());
        }
    }
}
